#! /usr/bin/env perl
use strict;
use warnings;
exit 0;

__END__

=head1 DESCRIPTION

This file serves only to test the detection of dependencies of perl scripts.
It should depend on the perl interpreter and strict.pm and warnings.pm and any
C libs the perl interpreter depends on.
